@@ -3,7 +3,7 @@ import os
 import shutil
 import unittest
 
-from mock import Mock, PropertyMock, patch
+from unittest.mock import Mock, patch
 
 import urllib3
 
@@ -20,15 +20,6 @@ class SleepException(Exception):
     pass
 
 
-mock_available_gucs = PropertyMock(return_value={
-    'cluster_name', 'constraint_exclusion', 'force_parallel_mode', 'hot_standby', 'listen_addresses', 'max_connections',
-    'max_locks_per_transaction', 'max_prepared_transactions', 'max_replication_slots', 'max_stack_depth',
-    'max_wal_senders', 'max_worker_processes', 'port', 'search_path', 'shared_preload_libraries',
-    'stats_temp_directory', 'synchronous_standby_names', 'track_commit_timestamp', 'unix_socket_directories',
-    'vacuum_cost_delay', 'vacuum_cost_limit', 'wal_keep_size', 'wal_level', 'wal_log_hints', 'zero_damaged_pages',
-    'autovacuum', 'wal_segment_size', 'wal_block_size', 'shared_buffers', 'wal_buffers',
-})
-
 GET_PG_SETTINGS_RESULT = [
     ('wal_segment_size', '2048', '8kB', 'integer', 'internal'),
     ('wal_block_size', '8192', None, 'integer', 'internal'),
@@ -66,7 +57,7 @@ class MockResponse(object):
 
     def __init__(self, status_code=200):
         self.status_code = status_code
-        self.headers = {'content-type': 'json'}
+        self.headers = {'content-type': 'json', 'lsn': 100}
         self.content = '{}'
         self.reason = 'Not Found'
 
@@ -138,9 +129,9 @@ class MockCursor(object):
         elif sql.startswith('SELECT slot_name, slot_type, datname, plugin, catalog_xmin'):
             self.results = [('ls', 'logical', 'a', 'b', 100, 500, b'123456')]
         elif sql.startswith('SELECT slot_name'):
-            self.results = [('blabla', 'physical', 12345),
-                            ('foobar', 'physical', 12345),
-                            ('ls', 'logical', 499, 'b', 'a', 5, 100, 500)]
+            self.results = [('blabla', 'physical', 1, 12345),
+                            ('foobar', 'physical', 1, 12345),
+                            ('ls', 'logical', 1, 499, 'b', 'a', 5, 100, 500)]
         elif sql.startswith('WITH slots AS (SELECT slot_name, active'):
             self.results = [(False, True)] if self.rowcount == 1 else []
         elif sql.startswith('SELECT CASE WHEN pg_catalog.pg_is_in_recovery()'):
@@ -179,8 +170,12 @@ class MockCursor(object):
                                  b'3\t0/403DD98\tno recovery target specified\n')]
         elif sql.startswith('SELECT pg_catalog.citus_add_node'):
             self.results = [(2,)]
-        elif sql.startswith('SELECT nodeid, groupid'):
-            self.results = [(1, 0, 'host1', 5432, 'primary'), (2, 1, 'host2', 5432, 'primary')]
+        elif sql.startswith('SELECT groupid, nodename'):
+            self.results = [(0, 'host1', 5432, 'primary', 1),
+                            (0, '127.0.0.1', 5436, 'secondary', 2),
+                            (1, 'host4', 5432, 'primary', 3),
+                            (1, '127.0.0.1', 5437, 'secondary', 4),
+                            (1, '127.0.0.1', 5438, 'secondary', 5)]
         else:
             self.results = [(None, None, None, None, None, None, None, None, None, None)]
         self.rowcount = len(self.results)
@@ -202,20 +197,16 @@ class MockCursor(object):
         pass
 
 
-class MockConnectionInfo(object):
-
-    def parameter_status(self, param_name):
-        if param_name == 'is_superuser':
-            return 'on'
-        return '0'
-
-
 class MockConnect(object):
 
     server_version = 99999
     autocommit = False
     closed = 0
-    info = MockConnectionInfo()
+
+    def get_parameter_status(self, param_name):
+        if param_name == 'is_superuser':
+            return 'on'
+        return '0'
 
     def cursor(self):
         return MockCursor(self)
